/**
 *
 * Copyright (C) 2013 Olivier Berthier
 *                    http://www.xunil.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

#ifndef POINTEUR_H
#define POINTEUR_H


#include <string>
#include <iostream>
#include <sstream>
#include <cstddef>
#include <new>
#include <vector>


class compteurPI;
template<typename T> 
class pIntelligent;
template<typename T> 
std::ostream& operator<<(std::ostream &, const pIntelligent<T> &);
template<typename T> 
class node;
template<typename T>
class vectorPI;



/**
 * \file pointeur.h
 * \brief Classe du compteur des pointeurs intelligents
 * \author Olivier BERTHIER
 * \date avril 2013
 *
 * Classe qui permet d'instancier un compteur de références pour un pointeur intelligent.
 */
class compteurPI{

 private: 
  /**
   * \brief Nombre de références
   */
  int nbr;

 public:
  /**
   * \brief Constructeur par défaut
   * 
   * Constructeur par défaut de la classe compteurPI, initialise le nombre de références à 0;
   */
  compteurPI() {
    this->nbr = 0;
#ifdef DEBUG
    std::cerr<<"["<<this<<"] compteurPI::new"<<std::endl;
#endif
  }

  /**
   * \brief Constructeur par copie
   *
   * Constructeur par copie de la classe compteurPI.
   *
   * \param const compteurPI&
   */
  compteurPI(const compteurPI& csp){ 
#ifdef DEBUG
    std::cerr<<"["<<this<<"] compteurPI::copy("<<&csp<<")"<<std::endl;
#endif
    this->nbr = csp.nbr; 
  }

  /**
   * \brief Opérateur d'affectation
   *
   * Surcharge de l'opérateur d'affectation de la classe compteurPI.
   *
   * \param const compteurPI&
   * \return compteurPI&
   */
  compteurPI& operator=(const compteurPI& csp){
#ifdef DEBUG
    std::cerr<<"["<<this<<"] compteurPI::copy("<<&csp<<")"<<std::endl;
#endif
    this->nbr = csp.nbr;
    return *this;
  }

  /**
   * \brief Destructeur
   *
   * Destructeur de la classe compteurPI.
   */
  ~compteurPI(){ }

  /**
   * \brief Procédure d'incrémentation du compteur
   *
   * Procédure permettant d'incrémenter le compteur de pointeurs intelligents.
   */
  void add(){
    nbr++; 
#ifdef DEBUG
    std::cerr<<"["<<this<<"] compteurPI::++"<<std::endl;
#endif
  }

  /**
   * \brief Procédure de décrémentation du compteur
   *
   * Procédure permettant de décrémenter le compteur de pointeurs intelligents.
   */
  void remove() {
    nbr--;
#ifdef DEBUG
    std::cerr<<"["<<this<<"] compteurPI::--"<<std::endl;
#endif
  }

  /**
   * \brief Retourne la valeur du compteur
   *
   * Fonction retournant la valeur du compteur.
   *
   * \return int
   */
  int count() const{ return nbr;}
};






/**
 * \file pointeur.h
 * \brief Classe template des noeuds associés aux pointeurs intelligents
 * \author Olivier BERTHIER
 * \date avril 2013
 *
 * Classe template permettant d'instancier des noeud associés aux pointeurs intelligents permettant de retrouver les parents et enfants d'un pointeur intelligent.
 */
template<typename T>
class node{
  
 private:
  
 protected:
  
  /**
   * \brief Pointeur vers un pointeur intelligent
   */
  pIntelligent<T>* ptrI;

  /**
   * \brief Vecteur de pointeurs sur les node parents
   */  
  std::vector<node<T>*> parents;

  /**
   * \brief Vecteur de pointeurs sur les node enfants
   */  
  std::vector<node<T>*> children;

  /**
   * \brief Etat du noeud
   *
   * 0 => pas encore vu
   * 1 => parent ok
   * 2 => noeud ok
   */
  int state;

 public:
  /**
   * \brief Constructeur par défaut
   *
   * Constructeur par défaut de la classe template node.
   */
  node(){}

  /**
   * \brief Constructeur
   * 
   * Constructeur de la classe template node.
   *
   * \param pIntelligent<T>* : Pointeur d'un pointeur intelligent
   */
  node(pIntelligent<T>* ptrI){
#ifdef DEBUG
    std::cerr<<"["<<this<<"] node::new"<<std::endl;
#endif
    this->ptrI = ptrI;
    this->state = 0;
  }

  /**
   * \brief Constructeur par copie
   *
   * Constructeur par copie de la classe template node.
   *
   * \param node<T>& : référence constante vers une instance de node
   */ 
  node(const node<T>& n){
#ifdef DEBUG
    std::cerr<<"["<<this<<"] node::copy("<<&n<<")"<<std::endl;
#endif
    this->ptrI = n.ptrI;
    this->state = n.state;
  }

  /**
   * \brief Destructeur de la classe template node
   *
   * Destructeur de la classe template node qui appelle la fonction clear().
   */
  ~node(){    
#ifdef DEBUG
    std::cerr<<"["<<this<<"] node::delete"<<std::endl;
#endif
    clear();   
  }

  /**
   * \brief Procédure de destruction des liens
   *
   * Procédure qui va retirer chez chacun de ses parents et enfants la référence vers elle ainsi que ses propres références.
   */
  void clear(){
    for(unsigned i = 0; i<parents.size();i++){ 
      parents[i]->removeChild(this);
#ifdef DEBUG
      std::cerr<<"["<<this<<"] node::removeParents(erease "<<&parents[i]<<") "<<parents[i]->ptrI->count()<<" refs "<<std::endl;
#endif
      parents.erase(parents.begin()+i);
    }

    for(unsigned i = 0; i<children.size();i++){ 
      children[i]->removeParent(this);
#ifdef DEBUG
      std::cerr<<"["<<this<<"] node::removeChild(erease "<<&children[i]<<") "<<children[i]->ptrI->count()<<" refs "<<std::endl;
#endif
      children.erase(children.begin()+i);
    }
  }

  /**
   * \brief Fonction de recherche de noeud enfant
   *
   * Fonction qui va rechercher si un noeud est présent dans ses enfants.
   *
   * \param const node<T>* : pointeur constant vers un noeud
   * \return bool : vrai si le noeud a été trouvé, faux sinon
   */
  bool findChild(const node<T>* n) const{
    for(unsigned i = 0; i<children.size();i++)
      if(children[i] == n) return true;      
    return false;  
  }
  /**
   * \brief Fonction de recherche de noeud parent
   *
   * Fonction qui va rechercher si un noeud est présent dans les parents du node.
   *
   * \param const node<T>* : pointeur constant vers un noeud
   * \return bool : vrai si le noeud a été trouvé, faux sinon
   */
  bool findParent(const node<T>* n) const{
    for(unsigned i = 0; i<parents.size();i++)
      if(parents[i] == n) return true;
    return false;  
  }

  /**
   * \brief Procédure d'ajout d'un noeud parent
   *
   * Procédure qui va ajouter le noeud passé en paramètre aux parents du noeud si celui-ci n'existe pas déjà, puis va ajouter lui même comme enfant du noeud parent.
   *
   * \param node<T>* : pointeur vers un noeud
   */  
  void addParent(node<T>* n){

    if(!findParent(n)){
      parents.push_back(n);
#ifdef DEBUG
      std::cerr<<"["<<this<<"] node::addParent("<<&n<<")"<<std::endl;
#endif
    }
    n->addChild(this);
  }
  

  /**
   * \brief Procédure de suppression d'un noeud parent
   *
   * Procédure qui va retirer aux parents du noeud le noeud passé en paramètre.
   *
   * \param node<T>* : pointeur vers un noeud
   */  
  void removeParent(node<T>* n){
#ifdef DEBUG
    std::cerr<<"["<<this<<"] node::removeParents("<<parents.size()<<")"<<std::endl;
#endif
    for(unsigned i = 0; i<parents.size();i++)
      if(parents[i] == n) 
	parents.erase(parents.begin()+i);
  }

  /**
   * \brief Procédure de suppression d'un noeud parent
   *
   * Procédure qui va retirer aux parents du noeud le noeud passé en paramètre.
   *
   * \param node<T>* : pointeur vers un noeud
   */  
  void removeChild(node<T>* n){
#ifdef DEBUG
    std::cerr<<"["<<this<<"] node::removeChild("<<children.size()<<")"<<std::endl;
#endif
    for(unsigned i = 0; i<children.size();i++) 
      if(children[i] == n) 
	children.erase(children.begin()+i);
  }
  
  /**
   * \brief Procédure de validation du noeud
   *
   * Procédure de validation du noeud : elle va assicier l'état 1 à chacun de ses enfants, puis passer à l'état 2 pour indiquer qu'elle est valide.
   */
  void check(){
    for(unsigned i=0; i<children.size(); i++)
      children[i]->state = 1;    
    state = 2;
  }

  /**
   * \brief Connaitre la présence d'enfants
   *
   * Fonction booléenne permettant de savoir si le noeud a des enfants.
   *
   * \return bool : vrai si le noeud à au moins 1 enfant, faux sinon
   */
  bool hasChild() const{ return children.size()>0; }

  /**
   * \brief Connaitre la présence de parents
   *
   * Fonction booléenne permettant de savoir si le noeud a des parents.
   *
   * \return bool : vrai si le noeud à au moins 1 parent, faux sinon
   */
  bool hasParent() const{ return parents.size()>0; }

  /**
   * \brief Retourne l'état du noeud
   *
   * Fonction qui retourne l'état du noeud :
   * 0 => pas encore vu
   * 1 => parent ok
   * 2 => noeud ok
   *
   * \return int : état du noeud
   */
  int getState() const{ return state; }

  /**
   * \brief Affect l'état du noeud
   *
   * Procédure qui affect l'état du noeud :
   * 0 => pas encore vu
   * 1 => parent ok
   * 2 => noeud ok
   *
   * \param int : état du noeud
   */
  void setState(int n){ state=n; }

  /**
   * \brief Opérateur d'affectation
   *
   * Opérateur d'affectation de la classe template node.
   * 
   * \param const node<T>& : référence constante vers un node
   * \return node<T>& : référence vers un node
   */
  node<T>& operator=(const node<T>& n){
#ifdef DEBUG
    std::cerr<<"["<<this<<"] node::affect("<<&n<<")"<<std::endl;
#endif
    this->ptrI = n.ptrI;
    this->state = n.state;
    return *this;
  }

  /**
   * \brief Opérateur d'égalité
   *
   * Opérateur d'égalité de la classe template node. La comparaison n'est faire que sur l'égalité des pointeurs intelligents.
   *
   * \param const node<T>& : référence constante vers un node
   * \return bool : vrai si le node et celui passé en paramètre sont égaux, faux sinon.
   */
  bool operator==(const node<T>& n) const {
    return &this->ptrI == &n.ptrI ;
  }

 protected:
  /**
   * \brief Procédure d'ajout d'un noeud enfant
   *
   * Procédure qui va ajouter le noeud passé en paramètre aux enfants du noeud si celui-ci n'existe pas déjà.
   *
   * \param node<T>* : pointeur vers un noeud
   */  
  void addChild(node<T>* n){
    if(!findChild(n)){
      children.push_back(n);
#ifdef DEBUG
      std::cerr<<"["<<this<<"] node::addChild("<<&n<<")"<<std::endl;
#endif
    }
  }

};


/**
 * \file pointeur.h
 * \brief Classe template des pointeurs intelligents
 * \author Olivier BERTHIER
 * \date avril 2013
 *
 * Classe template qui permet d'instancier des pointeurs intelligents de type T.
 */
template<typename T> class pIntelligent {
  template<typename> friend class vectorPI;
 protected:

 private:
  /**
   * \brief Pointeur vers la donnée
   */
  T* pointeur; 

  /**
   * \brief Pointeur vers le compteur associé
   */
  compteurPI* csp;

  /**
   * \brief node associé au pointeur intelligent
   */
  node<T> n;

  /**
   * \brief Conteneur static permettant de stocker tous les pointeurs intélligents de type T
   */
  static vectorPI<T> vRef;

 public:
  /**
   * \brief Consructeur par défaut
   *
   * Constructeur par défaut de la classe template pIntelligent :
   * - affecter le pointeur au pointeur intelligent
   * - initialiser le compteur
   * - initialiser le node et l'ajouter à vRef
   *
   * \param T* : pointeur vers une donnée de type T
   */
  pIntelligent(T* p=0) { 
    this->pointeur = p;
#ifdef DEBUG
    std::cerr<<"["<<this<<"] pIntelligent::new("<<&pointeur<<")"<<std::endl;
#endif
    csp = new compteurPI();
    csp->add();
    n = node<T>(this);
    vRef.push_back(this);
  } 

  /**
   * \brief Constructeur par copie
   *
   * Constructeur par copie de la classe template pIntelligent :
   * - le pointeur prend la référence du pointeur du pointeur intelligent passé en parametre
   * - décrémenter le compteur
   * - si le compteur atteint 0 détruire le compteur et la donnée
   * - associer le compteur du pointeur intelligent passé en paramètre au compteur de this
   * 
   * /param const pIntelligent<T>& 	
   */
  pIntelligent(const pIntelligent<T>& sp) {
#ifdef DEBUG
    std::cerr<<"["<<this<<"] pIntelligent::copy("<<&sp<<")"<<std::endl;
#endif
    this->pointeur = sp.pointeur;

    csp = sp.csp;
    csp->add();
    node<T>& tmp = const_cast<node<T>&>(sp.n);
    n.addParent(&tmp);
    
    //n = sp.n;
    //vRef.push_back(this);
  }

  /**
   * \brief Desctucteur
   *
   * Destructeur de la classe template pIntelligent :
   * - décrémenter le compteur
   * - si le compteur arrive à 0, détruire la donnée, le compteur et le node tout en le retirant de vRef 
   * - faire un check de l'ensemble des node
   *
   */
  ~pIntelligent() {
    vRef.check();
    csp->remove();          
#ifdef DEBUG
    std::cerr<<"["<<this<<"]  pIntelligent::delete ? "<<&csp<<"=>"<<csp->count()<<" ref"<<std::endl;
#endif
    if(csp->count()==0) {
      // Retirer les références du node vers fils et parents
      n.clear();
      if(pointeur!=NULL){
	delete pointeur;
	pointeur = NULL;
      }
      if(csp != NULL){
	delete csp;
	csp = NULL;
      }
#ifdef DEBUG
      std::cerr<<"["<<this<<"] pIntelligent::delete OK"<<std::endl;
#endif            
    }
    vRef.erase(this);
  }

  /**
   * \brief Procédure de nettoyage
   *
   * Procédure permettant de nettoyer le conteneur de pointeurs intelligents.
   */
  void forceDelete(){  
#ifdef DEBUG
    std::cerr<<"["<<this<<"]  pIntelligent::delete ? "<<&csp<<"=>"<<csp->count()<<" ref"<<std::endl;
#endif
    n.clear();
    csp->remove();
    if(csp->count()==0) {
      // Retirer les références du node vers fils et parents
      if(pointeur != NULL){       
	delete pointeur;
	pointeur = NULL;
      }
      if(csp != NULL){
	delete csp;
	csp = NULL;
      }
#ifdef DEBUG
      std::cerr<<"["<<this<<"] pIntelligent::delete OK"<<std::endl;
#endif      
     
    }
    vRef.erase(this);      
  }

  /**
   * \brief Opérateur affectation
   *
   * Opérateur d'affectation de la classe template pIntelligent :
   * - décrémenter le compteur 
   * - si le compteur arreint 0 effacer la donnée et le pointeur ainsi 
   * - effacer les références du node
   * - associer le compteur, pointeur 
   * - associer le node du pointeur intelligent passé en parametre comme parent au node de this
   * - incrémenter le compteur
   *
   * \param const pIntelligent<T>& : référence constante vers un pointeur intelligent
   * \return pIntelligent<T> 
   */
  pIntelligent<T>&  operator=(const pIntelligent<T>& sp) { 
#ifdef DEBUG
    std::cerr<<"["<<this<<"] pIntelligent::affect("<<&sp<<")"<<std::endl;
#endif
    n.clear();
    csp->remove(); 
    if(csp->count()==0) {
      // Retirer les références du node vers fils et parents
      n.clear();
      if(pointeur != NULL){    
	delete pointeur;
	pointeur = NULL;
      }      
      if(csp != NULL){
	delete csp;
	csp = NULL;
      }      
    }  
 
    csp = sp.csp;     
    pointeur = sp.pointeur;      

    node<T>& tmp = const_cast<node<T>&>(sp.n);
    n.addParent(&tmp);
    
    csp->add();      
        
    return *this; 
  }

  /**
   * \brief Surcharge opérateur new
   *
   * Pas de gestion de plage de mémoire -> non utilisé
   *
   * \param size_t
   * \return void*
   */
  void* operator new(size_t sz){
    void* m = malloc(sz);
    return m;
  }

  /**
   * \brief Surcharge opérateur delete
   *
   * Idem new(), non utilisé.
   * \param  void*
   */
  void operator delete(void* m) {
    free(m);
  }

  /**
   * \brief Opérateur d'égalité
   *
   * Opérateur d'égalité de la classe template pIntelligent. La comparaison n'est faite que sur le pointeur de la donnée.
   *
   * \param const pIntelligent<T>& : référence constante vers un pointeur intélligent
   * \return bool : vrai si les pointeurs des deux pointeurs intelligents sont les mêmes, faux sinon
   */
  bool  operator==(const pIntelligent<T>& sp) const { 
    return this->pointeur == sp.pointeur;
  }
  
  /**
   * \brief Surcharge de l'opérateur -> 
   *
   * \return T*
   */
  T* operator->()  { return pointeur; }

  /**
   * \brief Surcharge de l'opérateur -> 
   *
   * \return const T* 
   */
  const T* operator->() const { return pointeur; }
  
  /**
   * \brief Surcharge de l'opérateur * 
   *
   * \return T&
   */
  T& operator*() { return *pointeur; }

  /**
   * \brief Surcharge de l'opérateur *
   *
   * \return const T&
   */
  const T& operator*() const { return *pointeur; }

  
  /**
   * \brief Fonction toString
   *
   * Fonction qui retourne une chaine de caractere contenant le nombre de références (Il serait aussi possible de retourner la chaine de caractere de la donnée du pointeur intelligent, mais il faudrait être certain que l'opérateur "<<" soit implémenté)
   *
   * \return const std::string
   */
  const std::string toString() const { 
    std::ostringstream oss;
    oss << pointeur ;
    //    oss << " ( nb refs : "  << this->count() << " )";
    return std::string( oss.str());
  }

  /**
   * \brief Valeur du compteur
   *
   * Fonction qui retourne la valeur du compteur associé au pointeur intélligent.
   *
   * \return int
   */
  int count() const{ return csp->count(); }

  /**
   * \brief Surcharge de l'opérateur <<
   *
   * Fonction amie permettant de sortir le résultat de la fonction toString() dans un flux std::ostream.
   * 
   * \param std::ostream& : flux de sortie
   * \param const pIntelligent<T>& : référence constante vers un pointeur intelligent
   * \return std::ostream&
   */
  friend std::ostream & operator<< <>(std::ostream &, const pIntelligent<T> &);  
    
};






/**
 * \file pointeur.h
 * \brief Classe template du conteneur des noeuds associés aux pointeurs intelligents
 * \author Olivier BERTHIER
 * \date avril 2013
 *
 * Classe template qui permet d'instancier un conteneur de pointeurs intelligents, cela permet de faire un check sur l'ensemble des pointeurs intelligents et eventuellement forcer la libération de la mémoire lors de la destruction.
 */
template<typename T>
class vectorPI{
 protected:

 private:

  /**
   * \brief vecteur de pointeurs de pointeurs intelligents
   */
  std::vector<pIntelligent<T>*> v;

 public:
  /**
   * \brief Constructeur par défaut
   *
   * Constructeur par défaut de la classe template vectorPI
   */    
  vectorPI(){}

  /**
   * \brief Destructeur
   *
   * Destructeur de la classe template vectorPI :
   * - effacer toutes les références des pointeurs
   * - supprimer tous les elements du vecteur ainsi que les pointeurs intelligents associés
   */    
  ~vectorPI(){
    check();
    clear();
  }

  /**
   * \brief Opérateur []
   *
   * Surcharge de l'opérateur [].
   *
   * \param int : index
   * \return T& référence vers un objet de type T
   */
  T& operator[] (int i) {
    return v[i];
  }

  /**
   * \brief Opérateur []
   *
   * Surcharge de l'opérateur [].
   *
   * \param int : index
   * \return const T : un objet constant de type T
   */
  const T operator[] (int i) const {
    return v[i];
  }

  /**
   * \brief Procédure d'ajout au vecteur
   *
   * Procédure permettant d'ajouter un objet de type T.
   *
   * \param pIntelligent<T*> : pointeur intélligent de type T
   */
  void push_back(pIntelligent<T>* e){
    v.push_back(e);
  }

  /**
   * \brief Procédure d'effacement
   *
   * Procédure permettant d'éfacer un objet dans le conteneur.
   *
   * \param pIntelligent<T*> : pointeur intélligent de type T
   */
  void erase( pIntelligent<T>* e){
    for(unsigned i = 0; i<v.size(); i++)
      if(v[i]==e) v.erase(v.begin()+i);
  }

  /**
   * \brief Nombre d'éléments dans le conteneur
   *
   * Fonction qui retourne le nombre d'éléments dans le conteneur.
   *
   * \return int
   */
  int size(){
    return v.size();
  }

  /**
   * \brief Procédure pour éffacer les références circulaires
   *
   * - Mettre l'ensemble des node à 0
   * - Mettre à 1 chacun des node sans parents
   * - Tant qu'il y'a des node à 1 faire un check() sur le node (prendre chacun de leur enfants et les mettre à 1, puis mettre à 2 le node courant)
   * - Supprimer tous les node à 0
   */
  void check(){
    for(unsigned i=0; i<v.size(); i++){
      if(v[i]->n.hasParent() || v[i]->count() <=  0) v[i]->n.setState(0);
      else v[i]->n.setState(1);
    }

    bool valide;
    do{
      valide = true;
      for(unsigned i=0; i<v.size(); i++){
	if(v[i]->n.getState() == 1 ){	  
	  v[i]->n.check();
	  valide = false;
	}	  
      }	  
    }while(!valide);

    for(unsigned i=0; i<v.size(); i++){
      if(v[i]->n.getState() == 0 ){

	v[i]->n.clear();
      }
    } 
  }

  /**
   * \brief Forcer la destruction des pointeurs intelligents
   *
   * Procédure permettant de détruire tous les pointeurs intelligents (un peu violent).
   */
  void clear(){

    while(v.size()>0)
      v[0]->forceDelete();
        
    v.clear();
  }

};




template <typename T>
std::ostream & operator<<(std::ostream & output, const pIntelligent<T> & sp){
  output << sp.toString() << std::endl;
  return output;

}

      
//template<typename T> vectorPI<node<T>> pIntelligent<T>::vRef = vectorPI<node<T>>();
template<typename T> vectorPI<T> pIntelligent<T>::vRef = vectorPI<T>();


      
#endif
